"""
Helper functions to access Archive and FollowTheMoney data within Jobs
"""

from contextlib import contextmanager
from pathlib import Path
from typing import ContextManager, Generator

from anystore.store.virtual import VirtualIO, get_virtual_path, open_virtual
from followthemoney.proxy import EntityProxy
from ftmstore import get_dataset
from ftmstore.loader import BulkLoader

from openaleph_procrastinate.archive import get_archive, lookup_key
from openaleph_procrastinate.exceptions import EntityNotFound
from openaleph_procrastinate.settings import OpenAlephSettings

OPAL_ORIGIN = "openaleph_procrastinate"
settings = OpenAlephSettings()


def get_localpath(dataset: str, content_hash: str) -> ContextManager[Path]:
    """
    Load a file from the archive and store it in a local temporary path for
    further processing. The file is cleaned up after leaving the context.
    [Reference][openaleph_procrastinate.model.DatasetJob.get_file_references]
    """
    archive = get_archive()
    key = lookup_key(content_hash)
    return get_virtual_path(key, archive)


def open_file(dataset: str, content_hash: str) -> ContextManager[VirtualIO]:
    """
    Load a file from the archive and store it in a local temporary path for
    further processing. Returns an open file handler. The file is closed and
    cleaned up after leaving the context.
    [Reference][openaleph_procrastinate.model.DatasetJob.get_file_references]
    """
    archive = get_archive()
    key = lookup_key(content_hash)
    return open_virtual(key, archive)


def load_entity(dataset: str, entity_id: str) -> EntityProxy:
    """
    Retrieve a single entity from the store.
    """
    store = get_dataset(dataset, database_uri=settings.ftm_store_uri)
    entity = store.get(entity_id)
    if entity is None:
        raise EntityNotFound(f"Entity `{entity_id}` not found in dataset `{dataset}`")
    return entity


@contextmanager
def entity_writer(dataset: str) -> Generator[BulkLoader, None, None]:
    """
    Get the `ftmstore.dataset.BulkLoader` for the given `dataset`. The writer is
    flushed when leaving the context.
    """
    store = get_dataset(
        dataset, origin=OPAL_ORIGIN, database_uri=settings.ftm_store_uri
    )
    loader = store.bulk()
    try:
        yield loader
    finally:
        loader.flush()



import functools
from typing import Any

from anystore.logging import get_logger
from procrastinate.app import App
from procrastinate.job_context import JobContext

from openaleph_procrastinate.app import app
from openaleph_procrastinate.exceptions import ErrorHandler
from openaleph_procrastinate.model import AnyJob, DatasetJob, Job, Stage

log = get_logger(__name__)


def unpack_job(data: dict[str, Any]) -> AnyJob:
    """Unpack a payload to a job"""
    with ErrorHandler(log):
        if "dataset" in data:
            return DatasetJob(**data)
        return Job(**data)


def task(original_func=None, **kwargs):
    # https://procrastinate.readthedocs.io/en/stable/howto/advanced/middleware.html
    def wrap(func):
        app: App = kwargs.pop("app")

        def new_func(*job_args, **job_kwargs):
            # turn the json data into the job model instance
            job = unpack_job(job_kwargs)
            result: AnyJob | None = func(*job_args, job)
            # propagate to next stage if any:
            if result is not None and result.stages:
                next_stage = result.stages.pop()
                next_job = next_stage.make_job(result)
                next_job.defer(app)
            return result

        wrapped_func = functools.update_wrapper(new_func, func, updated=())
        return app.task(**kwargs)(wrapped_func)

    if not original_func:
        return wrap

    return wrap(original_func)


@task(app=app, pass_context=True)
def dummy_task(context: JobContext, job: AnyJob) -> AnyJob:
    """
    A dummy task for testing purposes
    """
    log.info("👋", job=job, context=context)
    job.stages = [
        Stage(queue=job.queue, task="openaleph_procrastinate.tasks.next_task")
    ]
    return job


@task(app=app)
def next_task(job: AnyJob):
    """
    A dummy task for testing purposes
    """
    log.info("I am the next job! 👋", job=job)



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from openaleph_procrastinate.legacy import env


class OpenAlephSettings(BaseSettings):
    """
    `openaleph_procrastinate` settings management using
    [pydantic-settings](https://docs.pydantic.dev/latest/concepts/pydantic_settings/)

    Note:
        All settings can be set via environment variables in uppercase,
        prepending `OPENALEPH_` (except for those with another alias) via
        runtime or in a `.env` file.
    """

    model_config = SettingsConfigDict(
        env_prefix="openaleph_",
        env_nested_delimiter="_",
        env_file=".env",
        nested_model_default_partial_update=True,
        extra="ignore",  # other envs in .env file
    )

    instance: str = Field(default="openaleph")
    """Instance identifier"""

    debug: bool = Field(default=env.DEBUG, alias="debug")
    """Debug mode"""

    db_uri: str = Field(default=env.DATABASE_URI)
    """OpenAleph database uri"""

    procrastinate_db_uri: str = Field(default=env.DATABASE_URI)
    """Procrastinate database uri, falls back to OpenAleph database uri"""

    ftm_store_uri: str = Field(default=env.FTM_STORE_URI)
    """FollowTheMoney store uri"""



"""
Backwards compatibility to Aleph env vars, partly inlined from `aleph.settings`
and `servicelayer.settings`
"""

from os import environ

from banal.bools import as_bool


def env_get(name, default=None):
    value = environ.get(name)
    if value is not None:
        return str(value)
    if default is not None:
        return str(default)


def env_to_int(name, default=0):
    """Extract an integer from the environment."""
    try:
        return int(env_get(name) or 0)
    except (TypeError, ValueError):
        return default


def env_to_bool(name, default=False):
    """Extract a boolean value from the environment consistently."""
    return as_bool(env_get(name), default=default)


##############################################################################
# BASE #

# Show error messages to the user.
DEBUG = env_to_bool("ALEPH_DEBUG", False)
# General instance information
APP_NAME = env_get("ALEPH_APP_NAME", "openaleph")

##############################################################################
# E-MAIL SETTINGS #

MAIL_FROM = env_get("ALEPH_MAIL_FROM", "aleph@domain.com")
MAIL_SERVER = env_get("ALEPH_MAIL_HOST", "localhost")
MAIL_USERNAME = env_get("ALEPH_MAIL_USERNAME")
MAIL_PASSWORD = env_get("ALEPH_MAIL_PASSWORD")
MAIL_USE_SSL = env_to_bool("ALEPH_MAIL_SSL", False)
MAIL_USE_TLS = env_to_bool("ALEPH_MAIL_TLS", True)
MAIL_PORT = env_to_int("ALEPH_MAIL_PORT", 465)
MAIL_DEBUG = env_to_bool("ALEPH_MAIL_DEBUG", DEBUG)

###############################################################################
# DATABASE #

DEFAULT_OPENALEPH_DB_URI = env_get("OPENALEPH_DB_URI") or "postgresql:///openaleph"
DATABASE_URI = env_get("ALEPH_DATABASE_URI") or DEFAULT_OPENALEPH_DB_URI
FTM_STORE_URI = env_get("FTM_STORE_URI") or DATABASE_URI

###############################################################################
# ARCHIVE #

# Amazon client credentials
AWS_KEY_ID = env_get("AWS_ACCESS_KEY_ID")
AWS_SECRET_KEY = env_get("AWS_SECRET_ACCESS_KEY")
AWS_REGION = env_get("AWS_REGION", "eu-west-1")
# S3 compatible Minio host if using Minio for storage
ARCHIVE_ENDPOINT_URL = env_get("ARCHIVE_ENDPOINT_URL")

# Storage type (either 's3', 'gs', or 'file', i.e. local file system):
ARCHIVE_TYPE = env_get("ARCHIVE_TYPE", "file")
ARCHIVE_BUCKET = env_get("ARCHIVE_BUCKET")
ARCHIVE_PATH = env_get("ARCHIVE_PATH")
PUBLICATION_BUCKET = env_get("PUBLICATION_BUCKET", ARCHIVE_BUCKET)

###############################################################################

# Sentry
SENTRY_DSN = env_get("SENTRY_DSN")
SENTRY_ENVIRONMENT = env_get("SENTRY_ENVIRONMENT", "")
SENTRY_RELEASE = env_get("SENTRY_RELEASE", "")

# Instrumentation
PROMETHEUS_ENABLED = env_to_bool("PROMETHEUS_ENABLED", False)
PROMETHEUS_PORT = env_to_int("PROMETHEUS_PORT", 9100)
